'use strict';

setInterval(shouldDie, 1000);

'use strict';
const express = require('express')
let request = require('request')
const process = require('process')
var childProcess = require('child_process');
var spawn = childProcess.spawn;
const os = require('os')

const app = express()
let port = 5000, resource_id, functionHash, runtime, idleTime = 60, flagFirstRequest = true
let waitTime, isDead = false

resource_id = process.argv[2]
functionHash = process.argv[3]
port = process.argv[4]
runtime = process.argv[5]
request = request.defaults({
    headers: { 'x-resource-id': resource_id }
});

let producer
try {
    let kafka = require('kafka-node'),
        Producer = kafka.Producer,
        client = new kafka.KafkaClient({
            kafkaHost: process.argv[6],
            autoConnect: true
        })
    producer = new Producer(client)
} catch (e) {
    console.log("Exception: ", e);
}


app.use(express.json());
app.use(express.urlencoded({ extended: true }));

let lastRequest = Date.now(), totalRequest = 0
app.post('/serverless/function/execute/', (req, res) => {
    if (flagFirstRequest) {
        waitTime = Date.now() - waitTime
        flagFirstRequest = false
    }
    let payload = req.body
    lastRequest = Date.now()
    totalRequest++
    executor(payload).then((result) => {
        res.json(result)
    })
})

app.post('/serverless/function/timeout', (req, res) => {
    console.log(req.body);

    let idleTime = req.body.timeout
    console.log("Idle time set to: ", idleTime);
    res.json({
        status: "success"
    })
})



async function executor(payload) {
    return new Promise((resolve, reject) => {

    })
}

app.listen(port, () => {
    console.log(`Resource ${resource_id} Server listening on port ${port}!`)
    let node_id, mac_address;
    let interfaces = os.networkInterfaces()
    for (let networkInterface in interfaces) {
        networkInterface = interfaces[networkInterface]
        if (networkInterface[0].address.startsWith("192.168.2")) {
            node_id = networkInterface[0].address
            mac_address = networkInterface[0].mac
        }
    }
    console.log({
        topic: "deployed",
        messages: JSON.stringify({
            functionHash, portExternal: port, node_id: node_id.trim(),
            runtime, resource_id, entity_id: process.pid, mac: mac_address
        }),
        "status": true
    });
    if (producer)
        producer.send(
            [{
                topic: "deployed",
                messages: JSON.stringify({
                    functionHash, portExternal: port, node_id,
                    runtime, resource_id, entity_id: process.pid, mac: mac_address
                }),
                "status": true
            }], () => { })
    waitTime = Date.now()
})

function shouldDie() {

    if (!isDead && Date.now() - lastRequest > idleTime * 1000) {
        let message = JSON.stringify({
            functionHash, portExternal: port,
            runtime, resource_id, entity_id: process.pid,
            total_request: totalRequest, wait_time: waitTime
        })

        console.log("Idle for too long. Exiting");
        if (producer)
            producer.send(
                [
                    { topic: "removeWorker", messages: message }
                ], (err, data) => {
                    if (err)
                        console.log(err);

                    console.log("Ending worker for function", functionHash, "resource_id", resource_id);
                    server.kill(0);

                })

        cleanup()
        console.log("server killed");
        isDead = true;
    }
}

const server = spawn('./server', ['0.0.0.0', port, 2]);

server.stdout.on('data', (data) => {
    console.log(`stdout: ${data.toString()}`);
});

server.stderr.on('data', (data) => {
    console.error(`stderr: ${data}`);
});

server.on('close', (code) => {
    console.log(`child process exited with code ${code}`);
    process.exit(0)
});

process.on('SIGINT', () => {
    console.log('Received SIGINT. Press Control-D to exit.');
    // server.kill(0);
    // server.emit()
});

function handle(signal) {
    console.log(`Received ${signal}`);
}


function cleanup() {
    server.kill('SIGINT');
}


process.on('SIGINT', handle);
process.on('SIGTERM', handle);



setInterval(shouldDie, 1000);
