'use strict';
const express = require('express')
let request = require('request')
const process = require('process')
require('events').EventEmitter.defaultMaxListeners = Infinity;

const dgram = require('dgram');
const { constants } = require('buffer');
const { spawnSync, execSync } = require('child_process');
// const { logger } = require('../../lib');
const server = dgram.createSocket('udp4');
const udpProxy = dgram.createSocket('udp4');
const os = require('os')
// const os_utils = require("os-utils")
let struct = require('jspack')
struct = struct.jspack

const app = express()
let port = 5000, resource_id, functionHash, runtime, idleTime = 60, flagFirstRequest = true
let waitTime
resource_id = process.argv[2]
functionHash = process.argv[3]
port = process.argv[4]
runtime = process.argv[5]
request = request.defaults({
    headers: { 'x-resource-id': resource_id }
});

let requestQueue = []

console.log("function started , info : ",port , resource_id, functionHash, runtime, idleTime , flagFirstRequest)

let producer
try {
    let kafka = require('kafka-node'),
        Producer = kafka.Producer,
        client = new kafka.KafkaClient({
            kafkaHost: process.argv[6],
            autoConnect: true
        })
    producer = new Producer(client)
} catch(e) {
    console.log("Exception: ", e);
}


app.use(express.json());
app.use(express.urlencoded({ extended: true }));

let lastRequest = Date.now(), totalRequest = 0, requestLoad = 0
app.post('/serverless/function/execute/', (req, res) => {
    if (flagFirstRequest) {
        waitTime = Date.now() - waitTime
        flagFirstRequest = false
    }
    let payload = req.body
    lastRequest = Date.now()
    totalRequest++
    executor(payload).then((result) => {
        res.json(result)
    })
})

app.post('/serverless/function/timeout', (req, res) => {
    console.log(req.body);
    
    let idleTime = req.body.timeout
    console.log("Idle time set to: ", idleTime);
    res.json({
        status: "success"
    })
})

const sleep = (waitTimeInMs) => new Promise(resolve => setTimeout(resolve, waitTimeInMs));

async function executor(payload) {
    // let s=  Date.now()
    // console.log("payload exec id : ", payload.exec_id, port)
    // await sleep(10000)

    // let f = Date.now()
    // console.log("time taken to sleep: ",f-s, port)
    console.log("payload exec id  : ", payload.exec_id, port)
    return new Promise((resolve, reject) => {
        
        requestLoad--

        resolve(payload)
    })
}

let node_id, mac_address = "00:22:22:22:22:22";

app.listen(port, () => {
    console.log(`Resource ${resource_id} Server listening on port ${port}!`)
    
    let interfaces = os.networkInterfaces()
    for (let networkInterface in interfaces) {
        networkInterface = interfaces[networkInterface]
        if (networkInterface[0].address.startsWith("192.168")) {
            node_id = networkInterface[0].address
            mac_address = networkInterface[0].mac
        }
    }
    // mac_address = constants
    console.log({
        topic: "deployed3",
        messages: JSON.stringify({
            functionHash, portExternal: port, node_id: node_id.trim(),
            runtime, resource_id, entity_id: process.pid, mac: mac_address
        }),
        "status": true
    });
    if (producer)
        producer.send(
        [{
            topic: "deployed3",
            messages: JSON.stringify({ functionHash, portExternal: port,  node_id,
                runtime, resource_id, entity_id: process.pid, mac: mac_address}),
            "status": true
        }], () => { })
    waitTime = Date.now()
})

function shouldDie() {
    
    if (Date.now() - lastRequest > idleTime * 1000) {
        let message = JSON.stringify({
            functionHash, portExternal: port,
            runtime, resource_id, entity_id: process.pid,
            total_request: totalRequest, wait_time: waitTime
        })

        console.log("Idle for too long. Exiting");
        if (producer)
            producer.send(
                [
                    {topic: "REMOVE_FUNCTION_INSTANCE2", messages: message }
                ], (err, data) => {
                    if (err)
                        console.log(err);
                        
                    console.log("Ending worker for function", functionHash, "resource_id", resource_id);
                    process.exit(0)
                })
        
    }
}


server.on('error', (err) => {
    console.log(`server error:\n${err.stack}`);
    server.close();
});

function getRandomInt(min, max) {
    min = Math.ceil(min);
    max = Math.floor(max);
    return Math.floor(Math.random() * (max - min + 1)) + min;
}

// function autoscaleFunction()
// {
//     let payload = [{
//         topic: "AUTOSCALE" ,
//         messages: JSON.stringify({
//                 "resource_id": resource_id,
//                 "timestamp": Date.now(),
//                 "node": 
//                     { 
//                         node_id: address, 
//                         port: port, 
//                         mac: mac
//                     },
//                 "function_info": {
//                     "function_hash": functionHash
//                 }
//         }),
//         partition: 0
//     }]
//     producer.send(payload, () => {
//         console.log(`Replied`)
//     })
// }

function function_heartbeat()
{
    let info = {
        'free_mem': os.freemem(),
        'cpu_count' : os.cpus().length,
        'total_mem' : os.totalmem(),
        'avg_load' : os.loadavg(),
        'cpu_usage' : 0
        // cpu_usage : os_utils.cpuUsage()
    }
    // os_utils.cpuUsage(function(value){
    //     info['cpu_usage'] = value*100
    // })
    let payload = [{
        topic: "FUNCTION_LOAD3" ,
        messages: JSON.stringify({
                resource_id,
                functionHash,
                "runtime": runtime,
                "timestamp": Date.now(),
                "node": 
                    { 
                        "node_id": node_id, 
                        "portExternal": port, 
                        "mac": mac_address
                    },
                "load_info": info,
                "open_request_count": requestLoad
            }),
            partition: 0
    }]
    // let payload = [{
    //     topic: "FUNCTION_LOAD3" ,
    //     messages: JSON.stringify({
    //             "resource_id": resource_id,
    //             "functionHash": functionHash,
    //             "runtime": runtime,
    //             "timestamp": Date.now(),
    //             "node": 
    //                 { 
    //                     "node_id": node_id, 
    //                     "portExternal": port, 
    //                     "mac": mac_address
    //                 },
    //             "load_info": info,
    //             "open_request_count": requestLoad
    //         }),
    //         partition: 0
    // }]

    // console.log("\n\n function load : ", payload)
    producer.send(payload, () => {
        console.log(`function load Replied`)
    })
}


server.on('message', (msg, rinfo) => {
    if (flagFirstRequest) {
        waitTime = Date.now() - waitTime
        flagFirstRequest = false
    }
    // console.log("message", msg)
    let payload = unpackPacket(msg)

    console.log(payload, typeof payload);
    lastRequest = Date.now()
    // console.log("network stack time", lastRequest - payload.t1)
    totalRequest++
    requestLoad++
    // test
    if(totalRequest > 10)
    {
        console.log("total request is greater than 10. So closing function!!")
        process.exit(0)
    }
    // test
    executor(payload).then(result => {
        result = packPacket(msg)
        let port = 10000 + getRandomInt(0, 10)
        try {
            udpProxy.send(msg, 0, msg.length, port, rinfo.address, function (err, bytes) {
                if (err)
                    console.log(err)
                // console.log("response via UDP")
            })
        } catch (e) {
            console.log(e)
        }

    })
    console.log("request load on function : ", requestLoad, "total request : ", totalRequest)
    // if(requestLoad > 5)
    // {
    //     console.log("\nautoscling function!!!\n");
    //     autoscaleFunction()
    // }
    // requestLoad--

});

function unpackPacket(packet) {
    // let buffer = new Array(1024)
    let chain_id = null, exec_id = null, function_count = null, function_id = null, data = null
    let base = 0, f0, f1, f2, f3, f4, t1, t2, t3, t4
    chain_id = struct.Unpack(">I", packet, base)
    base += 4
    exec_id = struct.Unpack(">I", packet, base)
    base += 4
    function_id = struct.Unpack(">I", packet, base)
    base += 4
    data = struct.Unpack(">I", packet, base)
    base += 4
    function_count = struct.Unpack("B", packet, base)
    base += 1

    f0 = struct.Unpack("B", packet, base)
    base += 1
    f1 = struct.Unpack("B", packet, base)
    base += 1
    f2 = struct.Unpack("B", packet, base)
    base += 1
    f3 = struct.Unpack("B", packet, base)
    base += 1
    f4 = struct.Unpack("B", packet, base)
    // base += 1

    // t1 = struct.Unpack("I", packet, base)
    // base += 8
    // t2 = struct.Unpack("I", packet, base)
    // base += 8
    // t3 = struct.Unpack("I", packet, base)
    // base += 8
    // t4 = struct.Unpack("I", packet, base)


    // console.log("chain_id", chain_id, "exec_id", exec_id, "data", data, "function_count", function_count, "function_id", function_id)


    return {
        chain_id: chain_id[0],
        exec_id: exec_id[0],
        data: data[0],
        function_count: function_count[0],
        function_id: function_id[0],
        f0, f1, f2, f3, f4
    }
}

function packPacket(dataPacket) {

    let message = new Array(1024)
    let base = 0, chain_id, exec_id, function_id, data, function_count
    chain_id = struct.PackTo(">I", message, base, [dataPacket.chain_id])
    base += 4
    exec_id = struct.PackTo(">I", message, base, [dataPacket.exec_id])
    base += 4
    function_id = struct.PackTo(">I", message, base, [dataPacket.function_id])
    base += 4
    data = struct.PackTo(">I", message, base, [dataPacket.data])
    base += 4
    function_count = struct.PackTo("B", message, base, [dataPacket.function_count])
    message = Buffer.from(message)
    return message
}


server.on('listening', () => {
    const address = server.address();
    console.log(`server listening ${address.address}:${address.port}`);
});

// server.bind(port, "192.168.2.3");
server.bind(port);

setInterval(shouldDie, 1000);
setInterval(function_heartbeat, 1000);
// setInterval(loadStat, 2000);
